/****************************************************************************
** Meta object code from reading C++ file 'udptracker.h'
**
** Created by: The Qt Meta Object Compiler version 63 (Qt 4.8.7)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../src/libktorrent-1.3.1/src/tracker/udptracker.h"
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'udptracker.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 63
#error "This file was generated using the moc from 4.8.7. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
static const uint qt_meta_data_bt__UDPTracker[] = {

 // content:
       6,       // revision
       0,       // classname
       0,    0, // classinfo
       7,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: signature, parameters, type, tag, flags
      16,   15,   15,   15, 0x08,
      50,   32,   15,   15, 0x08,
      92,   79,   15,   15, 0x08,
     136,   79,   15,   15, 0x08,
     195,  178,   15,   15, 0x08,
     221,  218,   15,   15, 0x08,
     262,   15,   15,   15, 0x08,

       0        // eod
};

static const char qt_meta_stringdata_bt__UDPTracker[] = {
    "bt::UDPTracker\0\0onConnTimeout()\0"
    "tid,connection_id\0connectReceived(Int32,Int64)\0"
    "tid,buf,size\0announceReceived(Int32,const Uint8*,Uint32)\0"
    "scrapeReceived(Int32,const Uint8*,Uint32)\0"
    "tid,error_string\0onError(Int32,QString)\0"
    "ar\0onResolverResults(net::AddressResolver*)\0"
    "manualUpdate()\0"
};

void bt::UDPTracker::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        Q_ASSERT(staticMetaObject.cast(_o));
        UDPTracker *_t = static_cast<UDPTracker *>(_o);
        switch (_id) {
        case 0: _t->onConnTimeout(); break;
        case 1: _t->connectReceived((*reinterpret_cast< Int32(*)>(_a[1])),(*reinterpret_cast< Int64(*)>(_a[2]))); break;
        case 2: _t->announceReceived((*reinterpret_cast< Int32(*)>(_a[1])),(*reinterpret_cast< const Uint8*(*)>(_a[2])),(*reinterpret_cast< Uint32(*)>(_a[3]))); break;
        case 3: _t->scrapeReceived((*reinterpret_cast< Int32(*)>(_a[1])),(*reinterpret_cast< const Uint8*(*)>(_a[2])),(*reinterpret_cast< Uint32(*)>(_a[3]))); break;
        case 4: _t->onError((*reinterpret_cast< Int32(*)>(_a[1])),(*reinterpret_cast< const QString(*)>(_a[2]))); break;
        case 5: _t->onResolverResults((*reinterpret_cast< net::AddressResolver*(*)>(_a[1]))); break;
        case 6: _t->manualUpdate(); break;
        default: ;
        }
    }
}

const QMetaObjectExtraData bt::UDPTracker::staticMetaObjectExtraData = {
    0,  qt_static_metacall 
};

const QMetaObject bt::UDPTracker::staticMetaObject = {
    { &Tracker::staticMetaObject, qt_meta_stringdata_bt__UDPTracker,
      qt_meta_data_bt__UDPTracker, &staticMetaObjectExtraData }
};

#ifdef Q_NO_DATA_RELOCATION
const QMetaObject &bt::UDPTracker::getStaticMetaObject() { return staticMetaObject; }
#endif //Q_NO_DATA_RELOCATION

const QMetaObject *bt::UDPTracker::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->metaObject : &staticMetaObject;
}

void *bt::UDPTracker::qt_metacast(const char *_clname)
{
    if (!_clname) return 0;
    if (!strcmp(_clname, qt_meta_stringdata_bt__UDPTracker))
        return static_cast<void*>(const_cast< UDPTracker*>(this));
    return Tracker::qt_metacast(_clname);
}

int bt::UDPTracker::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = Tracker::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 7)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    return _id;
}
QT_END_MOC_NAMESPACE
