/****************************************************************************
** Meta object code from reading C++ file 'peerdownloader.h'
**
** Created by: The Qt Meta Object Compiler version 63 (Qt 4.8.7)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../src/libktorrent-1.3.1/src/peer/peerdownloader.h"
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'peerdownloader.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 63
#error "This file was generated using the moc from 4.8.7. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
static const uint qt_meta_data_bt__PeerDownloader[] = {

 // content:
       6,       // revision
       0,       // classname
       0,    0, // classinfo
       6,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: signature, parameters, type, tag, flags
      24,   20,   19,   19, 0x0a,
      42,   20,   19,   19, 0x0a,
      58,   19,   19,   19, 0x0a,
      70,   20,   19,   19, 0x0a,
      90,   19,   19,   19, 0x0a,
      99,   19,   19,   19, 0x08,

       0        // eod
};

static const char qt_meta_stringdata_bt__PeerDownloader[] = {
    "bt::PeerDownloader\0\0req\0download(Request)\0"
    "cancel(Request)\0cancelAll()\0"
    "onRejected(Request)\0update()\0"
    "peerDestroyed()\0"
};

void bt::PeerDownloader::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        Q_ASSERT(staticMetaObject.cast(_o));
        PeerDownloader *_t = static_cast<PeerDownloader *>(_o);
        switch (_id) {
        case 0: _t->download((*reinterpret_cast< const Request(*)>(_a[1]))); break;
        case 1: _t->cancel((*reinterpret_cast< const Request(*)>(_a[1]))); break;
        case 2: _t->cancelAll(); break;
        case 3: _t->onRejected((*reinterpret_cast< const Request(*)>(_a[1]))); break;
        case 4: _t->update(); break;
        case 5: _t->peerDestroyed(); break;
        default: ;
        }
    }
}

const QMetaObjectExtraData bt::PeerDownloader::staticMetaObjectExtraData = {
    0,  qt_static_metacall 
};

const QMetaObject bt::PeerDownloader::staticMetaObject = {
    { &PieceDownloader::staticMetaObject, qt_meta_stringdata_bt__PeerDownloader,
      qt_meta_data_bt__PeerDownloader, &staticMetaObjectExtraData }
};

#ifdef Q_NO_DATA_RELOCATION
const QMetaObject &bt::PeerDownloader::getStaticMetaObject() { return staticMetaObject; }
#endif //Q_NO_DATA_RELOCATION

const QMetaObject *bt::PeerDownloader::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->metaObject : &staticMetaObject;
}

void *bt::PeerDownloader::qt_metacast(const char *_clname)
{
    if (!_clname) return 0;
    if (!strcmp(_clname, qt_meta_stringdata_bt__PeerDownloader))
        return static_cast<void*>(const_cast< PeerDownloader*>(this));
    return PieceDownloader::qt_metacast(_clname);
}

int bt::PeerDownloader::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = PieceDownloader::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 6)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    return _id;
}
QT_END_MOC_NAMESPACE
